# Lint as: python2, python3
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

#Trainer executor for AI-platform-trening
from tfx.dsl.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer import executor as ai_platform_trainer_executor
from tfx.proto import trainer_pb2

#model resolver
from tfx.dsl.experimental import latest_blessed_model_resolver
from tfx.types import Channel
from tfx.types.standard_artifacts import Model
from tfx.types.standard_artifacts import ModelBlessing

#Evaluator
import tensorflow_model_analysis as tfma

#Pusher
from tfx.proto import pusher_pb2
from tfx.extensions.google_cloud_ai_platform.pusher import executor as ai_platform_pusher_executor

"""TFX taxi template configurations.

This file defines environments for a TFX taxi pipeline.
"""

import os  # pylint: disable=unused-import
import re

# Navnet brukes for å identifisere denne pipelinen. Navnet brukes konkret i KFP og i GCS.
PIPELINE_NAME = 'dev-pipeline'

# Brukes for å definere config for cluster.
# 'standard' or 'autopilot'
CLUSTER_SCALING = 'standard' # Kan også bruke autopilot, sett da "autopilot"

# GCP related configs.

# Following code will retrieve your GCP project. You can choose which project
# to use by setting GOOGLE_CLOUD_PROJECT environment variable.
try:
  import google.auth  # pylint: disable=g-import-not-at-top
  try:
    _, GOOGLE_CLOUD_PROJECT = google.auth.default()
  except google.auth.exceptions.DefaultCredentialsError:
    GOOGLE_CLOUD_PROJECT = ''
except ImportError:
  GOOGLE_CLOUD_PROJECT = ''

# Specify your GCS bucket name here. You have to use GCS to store output files
# when running a pipeline with Kubeflow Pipeline on GCP or when running a job
# using Dataflow. Default is '<gcp_project_name>-kubeflowpipelines-default'.
# This bucket is created automatically when you deploy KFP from marketplace.

# Using pipeline name to name GCS bucket
bucket_name = PIPELINE_NAME

# String to lowercase
bucket_name = bucket_name.lower()

# For norwegian letters, use ae, o, a
bucket_name = bucket_name.replace('æ', 'ae')
bucket_name = bucket_name.replace('ø', 'o')
bucket_name = bucket_name.replace('å', 'a')

# Removing all characters but a-z, 0-9, whitespace, '_', '-' and '.'
bucket_name = re.sub("[^a-z0-9\s\_\-\.]+", "", bucket_name)

# Swapping whitespace with '-'
bucket_name = re.sub("\s", "-", bucket_name)

# Remove repeating '-' (dashes) and dash after '.' and '_'
iter = list(bucket_name)
bucket_name = ''
for i in range (0,len(iter)):
    if iter[i] == '-' and iter[i-1] in '.-_':
        iter[i] = ''
    bucket_name += str(iter[i])

GCS_BUCKET_NAME = GOOGLE_CLOUD_PROJECT + '-' + bucket_name

# TFX pipeline produces many output files and metadata. All output data will be
# stored under this OUTPUT_DIR.
OUTPUT_DIR = os.path.join('gs://', GCS_BUCKET_NAME, 'output')
INPUT_DIR = os.path.join('gs://', GCS_BUCKET_NAME, 'input')

# The last component of the pipeline, "Pusher" will produce serving model under
# SERVING_MODEL_DIR.
SERVING_MODEL_DIR = os.path.join(OUTPUT_DIR, 'serving_model')

# Specifies data file directory. DATA_PATH should be a directory containing CSV
# files for CsvExampleGen in this example. By default, data files are in the
# GCS path: `gs://{GCS_BUCKET_NAME}/input/data/`.

DATA_PATH = os.path.join(INPUT_DIR, 'data')
PREPROCESSING_PATH = os.path.join(INPUT_DIR, 'modules/preprocessing.py')
TRAINER_PATH = os.path.join(INPUT_DIR, 'modules/trainer.py')

# Setter region til europa. Pga begrensninger i GCP, kan vi ikke kjøre alle tjenester i samme region. Deploying av modeller vil skje i europe-west1, mens resten skjer i europe-west4.
GOOGLE_CLOUD_REGION = 'europe-west4'
PUSHER_REGION = 'europe-west1'

ENABLE_CACHE = True

#Argumenter til dataflow
DATAFLOW_BEAM_PIPELINE_ARGS = [
   '--project=' + GOOGLE_CLOUD_PROJECT,
   '--runner=DataflowRunner',
   '--temp_location=' + os.path.join('gs://', GCS_BUCKET_NAME, 'tmp'),
   '--region=' + GOOGLE_CLOUD_REGION,

   # Temporary overrides of defaults.
   '--disk_size_gb=50',
   '--experiments=shuffle_mode=auto',
   '--machine_type=e2-standard-8',
   ]

#Trening på GCP
GCP_AI_PLATFORM_TRAINING_ARGS = {
    'project': GOOGLE_CLOUD_PROJECT,
    'region': GOOGLE_CLOUD_REGION,
}

#Serving på GCP
GCP_AI_PLATFORM_SERVING_ARGS = {
    'model_name': PIPELINE_NAME.replace('-','_'),  # '-' is not allowed.
    'project_id': GOOGLE_CLOUD_PROJECT,
    'regions': [PUSHER_REGION],
}

#Parametere til TFX-komponenter

CSV_EXAMPLE_GEN_ARGS = {
    'input_base': DATA_PATH,
}

SCHEMA_GEN_ARGS = {
    'infer_feature_shape': True,
}

TRANSFORM_ARGS = {
    #sti til konfigurasjonsfil for transform (foerprosessering)
    'module_file': PREPROCESSING_PATH,
}

#treningssteg
TRAIN_NUM_STEPS = 100

#evalueringssteg
TRAINER_EVAL_NUM_STEPS = 15

TRAINER_ARGS = {
    #sti til konfigurasjonsfil for trainer
    'module_file': TRAINER_PATH,

    #får trening til å skje på GCP
    'custom_executor_spec': executor_spec.ExecutorClassSpec(ai_platform_trainer_executor.GenericExecutor),
    'custom_config': {
        ai_platform_trainer_executor.TRAINING_ARGS_KEY: GCP_AI_PLATFORM_TRAINING_ARGS
    },
    'train_args': trainer_pb2.TrainArgs(num_steps=TRAIN_NUM_STEPS),
    'eval_args': trainer_pb2.EvalArgs(num_steps=TRAINER_EVAL_NUM_STEPS)
}

MODEL_RESOLVER_ARGS = {
    'instance_name': 'latest_blessed_model_resolver',
    'resolver_class': latest_blessed_model_resolver.LatestBlessedModelResolver,
    'model': Channel(type=Model),
    'model_blessing': Channel(type=ModelBlessing),
}

# Change this value according to your use cases.
EVAL_ACCURACY_THRESHOLD = 0.6

# Uses TFMA to compute a evaluation statistics over features of a model and
# perform quality validation of a candidate model (compared to a baseline).
EVAL_CONFIG = tfma.EvalConfig(
    model_specs=[
        # This assumes a serving model with signature 'serving_default'. If
        # using estimator based EvalSavedModel, add signature_name: 'eval' and
        # remove the label_key.
        tfma.ModelSpec(label_key='tips')
    ],
    metrics_specs=[
        tfma.MetricsSpec(
            # The metrics added here are in addition to those saved with the
            # model (assuming either a keras model or EvalSavedModel is used).
            # Any metrics added into the saved model (for example using
            # model.compile(..., metrics=[...]), etc) will be computed
            # automatically.
            # To add validation thresholds for metrics saved with the model,
            # add them keyed by metric name to the thresholds map.
            metrics=[
                tfma.MetricConfig(class_name='ExampleCount'),
                tfma.MetricConfig(class_name='BinaryAccuracy',
                  threshold=tfma.MetricThreshold(
                      value_threshold=tfma.GenericValueThreshold(
                          lower_bound={'value': EVAL_ACCURACY_THRESHOLD}),
                      change_threshold=tfma.GenericChangeThreshold(
                          direction=tfma.MetricDirection.HIGHER_IS_BETTER,
                          absolute={'value': -1e-10})))
            ]
        )
    ],
    slicing_specs=[
        # An empty slice spec means the overall slice, i.e. the whole dataset.
        tfma.SlicingSpec(),
        # Data can be sliced along a feature column. In this case, data is
        # sliced along feature column trip_start_hour.
        tfma.SlicingSpec(feature_keys=['trip_start_hour'])
    ]
)

EVALUATOR_ARGS = {
    'eval_config': EVAL_CONFIG,
}

PUSHER_ARGS = {
    'push_destination': pusher_pb2.PushDestination(
        filesystem=pusher_pb2.PushDestination.Filesystem(
            base_directory=SERVING_MODEL_DIR
        )
    ),
    'custom_executor_spec': executor_spec.ExecutorClassSpec(ai_platform_pusher_executor.Executor),
    'custom_config': {
        #får serving til å skje på GCP
            ai_platform_pusher_executor.SERVING_ARGS_KEY:
                GCP_AI_PLATFORM_SERVING_ARGS
        },
}
